/*
 * This file is part of the ALT project.
 * Copyright (C) 2021 Slava Aseev <ptrnine@altlinux.org>
 *
 * This is free software; you can use this application under the GNU
 * General Public License, version 3. See the file "LICENSE" for the
 * exact licensing terms.
 */

#define _GNU_SOURCE

#include <dirent.h>
#include <errno.h>
#include <pwd.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <X11/Xauth.h>
#include <X11/Xlib.h>
#include <systemd/sd-bus.h>

#define HOSTNAME1_PATH "/org/freedesktop/hostname1"
#define HOSTNAME1_DEST "org.freedesktop.hostname1"

#define XAUTH_LOCK_RETRIES  10
#define XAUTH_LOCK_TIMEOUT  2
#define XAUTH_LOCK_DEADTIME 600

#define MAX_DISPLAY_NAME_LEN      32
#define REMOTE_DISPLAY_START_PORT 6000

#define DEFAULT_XAUTH_SUFFIX "/.Xauthority"

#if (__GLIBC__ == 2 && __GLIBC_MINOR__ >= 34) || (__GLIBC__ > 2)
    #include <linux/close_range.h>
    #define CLOSE_ALL() close_range(0, UINT_MAX, 0)
#else
    #define CLOSE_ALL()                                                                                                \
        do {                                                                                                           \
            for (int i = 0; i < sysconf(_SC_OPEN_MAX); ++i) close(i);                                                  \
        } while (0)
#endif

#ifdef ENABLE_DEBUG_OUTPUT
#define print_debug(...) do { printf(__VA_ARGS__); fflush(stdout); } while (0)
#else
#define print_debug(...)
#endif

#define print_error(...) do { fprintf(stderr, __VA_ARGS__); fflush(stderr); } while (0)

/* For defining variables with cleanup attributes */
#define autocleanup(TYPE) TYPE __attribute__ ((__cleanup__(TYPE##_unrefp)))

static void Display_unrefp(Display **display) {
    if (*display)
        XCloseDisplay(*display);
}

/* Represents an element/node in the display list */
typedef struct display_node {
    char                 display_name[MAX_DISPLAY_NAME_LEN + 1];
    int                  is_local;
    struct display_node *next;
} display_node;

/* Appends a display name to the display node list.
 * If *head is NULL - assign the allocated node to them.
 *
 * display_name_size can be < 0 (in this case display_name should be NULL-terminated)
 *
 * Returns the current allocated node (last node in the list)
 */
display_node *append_display(
    display_node **head, display_node *tail, const char *display_name, ssize_t display_name_size, int is_local) {
    display_node *node = malloc(sizeof(display_node));
    memset(node, 0, sizeof(display_node));

    strncpy(node->display_name,
            display_name,
            display_name_size < 0 ? MAX_DISPLAY_NAME_LEN : MIN((size_t)display_name_size, MAX_DISPLAY_NAME_LEN));
    node->is_local = is_local;
    node->next     = NULL;

    print_debug("Append new display %s\n", node->display_name);

    if (!*head)
        *head = node;
    else if (tail)
        tail->next = node;
    return node;
}

/* Erases display list, assigns NULL to *node.
 * The "*node" is allowed to be NULL.
 */
void free_display_list(display_node **node) {
    if (!*node)
        return;

    display_node *e = *node;
    *node           = NULL;

    while (e) {
        display_node *next = e->next;
        free(e);
        e = next;
    }
}

/* Checks if the display list contains specified display_name (number).
 * Returns true if contains, otherwise returns false.
 * The "list" argument is allowed to be NULL.
 */
bool display_list_contains(display_node *list, const char *display_name, size_t display_name_len) {
    for (display_node *display = list; display; display = display->next)
        if (strncmp(display->display_name, display_name, display_name_len) == 0)
            return true;
    return false;
}

/* Returns pointer to the specified column in the input string.
 * If specified columnt does not exists the function returns NULL.
 * The functions uses ' ' and '\t' as delimiters.
 *
 * NULL-aware (line is allowed to be NULL)
 */
char *take_column(char *line, size_t column_index) {
    if (!line)
        return NULL;

    char *p = line;

    while (*p == ' ' || *p == '\t')
        ++p;

    size_t column = 0;
    while (column != column_index && *p != '\0') {
        while (*p != ' ' && *p != '\t' && *p != '\0')
            ++p;
        while (*p == ' ' || *p == '\t')
            ++p;
        ++column;
    }

    if (*p == '\0')
        return NULL;
    return p;
}

/* Attempts to collect displays for the current user's sessions.
 * The resulted list is assumed to contain unexisted displays names.
 * This should not affects anything, because non-existed displays will not be written to the XAUTHORITY file anyway.
 *
 * Returned display list should be released with free_display_list().
 */
display_node *determine_displays() {
    display_node *result = NULL;
    display_node *tail   = NULL;

    /* Determine local display.
     * Use DISPLAY environment variable.
     */
    char *local_display_name = getenv("DISPLAY");
    if (local_display_name) {
        local_display_name = strchr(local_display_name, ':');
        if (local_display_name) {
            ++local_display_name;

            ssize_t size = -1;
            char   *end  = strchr(local_display_name, '.');
            if (end)
                size = end - local_display_name;

            tail = append_display(&result, tail, local_display_name, size, true);
        }
    }
    else {
        print_error("Environment variable $DISPLAY is not set\n");
    }

    /* Determine remote displays.
     * Search for records in /proc/net/tcp which local address has port > 6000.
     * The supposed display name calculates as: display_name = port - 6000
     * Consider records only with the current user UID.
     */
    FILE     *nettcp       = fopen("/proc/net/tcp", "r");
    const int max_line_len = 256;
    char      line[max_line_len];
    if (nettcp) {
        /* Skip the first line with headings */
        char *p = fgets(line, max_line_len, nettcp);
        (void)(p);

        while (fgets(line, max_line_len, nettcp)) {
            char *ipaddr_str = take_column(line, 1);
            char *uid_str    = take_column(ipaddr_str, 6);

            if (!ipaddr_str || !uid_str)
                continue;

            const char *port_str = strchr(ipaddr_str, ':');
            if (!port_str)
                continue;
            ++port_str;

            unsigned long port = strtoul(port_str, NULL, 16);
            if (port <= REMOTE_DISPLAY_START_PORT)
                continue;

            uid_t uid = (uid_t)atoi(uid_str);
            if (uid != getuid())
                continue;

            char display_name[MAX_DISPLAY_NAME_LEN + 1];
            snprintf(display_name, MAX_DISPLAY_NAME_LEN, "%lu", port - REMOTE_DISPLAY_START_PORT);
            tail = append_display(&result, tail, display_name, -1, false);
        }

        fclose(nettcp);
    }

    return result;
}

/* Represents an element/node in the auth list */
typedef struct xauth_node {
    Xauth             *auth;
    struct xauth_node *next;
} xauth_node;

/* Appends an Xauth node to the auth list.
 * If *head is NULL - assign the allocated node to them.
 * Returns the current allocated node (last node in the list).
 */
xauth_node *append_auth(xauth_node **head, xauth_node *tail, Xauth *element) {
    xauth_node *node = malloc(sizeof(display_node));
    node->auth       = element;
    node->next       = NULL;

    if (!*head)
        *head = node;
    else if (tail)
        tail->next = node;
    return node;
}

/* Erases auth list, assigns NULL to the *node
 * The "*node" is allowed to be NULL.
 */
void free_auth_list(xauth_node **node) {
    if (!*node)
        return;

    xauth_node *e = *node;
    *node         = NULL;

    while (e) {
        xauth_node *next = e->next;
        XauDisposeAuth(e->auth);
        free(e);
        e = next;
    }
}

/* Read all Xauth entries from file on "auth_file_path" into auth list.
 *
 * NULL-aware (auth_file_path is allowed to be NULL)
 */
xauth_node *build_auth_list(const char *auth_file_path) {
    if (!auth_file_path)
        return NULL;

    Xauth      *entry = NULL;
    xauth_node *head  = NULL;
    xauth_node *tail  = NULL;

    FILE *auth_file = fopen(auth_file_path, "rb");

    if (auth_file) {
        print_debug("Build auth list for %s\n", auth_file_path);
        while ((entry = XauReadAuth(auth_file)))
            tail = append_auth(&head, tail, entry);
        fclose(auth_file);
    }
    else {
        print_error("Can't open authority file %s\n", auth_file_path);
    }

    return head;
}

/* Write auth list to the file with specified auth_file_path.
 *
 * Returns true if the file was successfully written, otherwise returns false.
 */
bool commit_auth_list(const char *auth_file_path, xauth_node *auth_list) {
    bool ok = false;

    /* Write to the temp file */
    size_t path_size = strlen(auth_file_path);
    char   tmppath[path_size + sizeof(".tmp.write")];
    strcpy(tmppath, auth_file_path);
    strcpy(tmppath + path_size, ".tmp.write");

    FILE *tmpfile = fopen(tmppath, "wb");
    if (tmpfile) {
        print_debug("Rewrite authority file %s\n", auth_file_path);
        ok = true;

        for (xauth_node *e = auth_list; e; e = e->next) {
            if (!XauWriteAuth(tmpfile, e->auth)) {
                print_error("XauWriteAuth() failed\n");
                ok = false;
            }
        }

        if (fsync(fileno(tmpfile))) {
            print_error("fsync() failed with errno=%i\n", errno);
            ok = false;
        }

        if (fclose(tmpfile)) {
            print_error("fclose() failed with errno=%i\n", errno);
            ok = false;
        }
    }

    /* Replace the target */
    if (ok && XauLockAuth(auth_file_path, XAUTH_LOCK_RETRIES, XAUTH_LOCK_TIMEOUT, XAUTH_LOCK_DEADTIME) == LOCK_SUCCESS) {
        if (rename(tmppath, auth_file_path)) {
            print_error("rename() failed with errno=%i\n", errno);
            ok = false;
        }

        XauUnlockAuth(auth_file_path);
    }

    print_debug("commit_auth_list(): return %s\n", ok ? "true" : "false");
    return ok;
}

/* Replaces address (hostname) in records where all conditions successed:
 *   - .family == FamilyLocal
 *   - .addres == old_hostname
 *   - .number is in displays list
 *
 * Returns true if at least one hostname was replaced, otherwise returns false
 */
bool replace_auth_record_hostname(xauth_node   *auth_list,
                                 display_node *displays,
                                 const char   *old_hostname,
                                 const char   *new_hostname) {
    bool ok = false;
    size_t address_len = strlen(old_hostname);

    for (xauth_node *e = auth_list; e; e = e->next) {
        Xauth *a = e->auth;
        if (a->family == FamilyLocal && a->address_length == address_len &&
            display_list_contains(displays, a->number, a->number_length) &&
            memcmp(a->address, old_hostname, address_len) == 0) {

            unsigned short new_name_len = strlen(new_hostname);
            if (a->address_length < new_name_len) {
                free(a->address);
                a->address = (char *)malloc(new_name_len);
            }

            print_debug("Replace %s/unix:%.*s by %s/unix:%.*s\n",
                        old_hostname,
                        a->number_length,
                        a->number,
                        new_hostname,
                        a->number_length,
                        a->number);

            memcpy(a->address, new_hostname, new_name_len);
            a->address_length = new_name_len;
            ok = true;
        }
    }

    print_debug("replace_auth_record_hostname(): return %s\n", ok ? "true" : "false");
    return ok;
}

/* Returns ${HOME}/.Xauthority path or NULL if getpwuid() returns NULL (HOME comes from getpwuid(getuid())->pw_dir)
 *
 * The result string must be released with free()
 */
char *standard_xauthority_path() {
    struct passwd *pw = getpwuid(getuid());
    if (!pw) {
        int err = errno;
        print_error("getpwuid() failed, rc: %i\n", err);
        return NULL;
    }
    const char *homedir         = pw->pw_dir;
    size_t      homedir_len     = strlen(homedir);
    char       *xauthority_path = malloc(homedir_len + sizeof(DEFAULT_XAUTH_SUFFIX));
    strcpy(xauthority_path, homedir);
    strcpy(xauthority_path + homedir_len, DEFAULT_XAUTH_SUFFIX);

    return xauthority_path;
}

/* Returns "${TMPDIR}/xauth-${UID}-_" + display_num
 * (see kdeinit/kinit.cpp:setupX() for details)
 *
 * Returns NULL if TMPDIR is not set.
 *
 * The result string must be released with free()
 */
char *additional_xauthority_path(const char *display_num) {
    char *tmpdir = getenv("TMPDIR");
    if (!tmpdir) {
        print_error("Environment variable TMPDIR is not set\n");
        return NULL;
    }

    char uidstr[16];
    snprintf(uidstr, sizeof(uidstr), "%i", getuid());

    size_t tmpdir_len      = strlen(tmpdir);
    size_t uid_len         = strlen(uidstr);
    size_t display_num_len = strlen(display_num);
    size_t len             = tmpdir_len + (sizeof("/xauth-") - 1) + uid_len + (sizeof("-_") - 1) + display_num_len;

    char  *result = malloc(len + 1);
    size_t pos    = 0;

    strcpy(result, tmpdir);
    pos += tmpdir_len;
    strcpy(result + pos, "/xauth-");
    pos += sizeof("/xauth-") - 1;
    strcpy(result + pos, uidstr);
    pos += uid_len;
    strcpy(result + pos, "-_");
    pos += sizeof("-_") - 1;
    strcpy(result + pos, display_num);

    print_debug("Additional xauthority path: %s\n", result);
    return result;
}

/* Replaces hostnames in $HOME/.Xauthority, and in XauFileName() if it differs from $HOME/.Xauthority
 * Returns true if at least one hostname replacement occured and successfully wrote in authority file,
 * otherwise returns false.
 */
bool perform_auth_changes(const char *old_hostname, const char *new_hostname) {
    bool ok = false;

    char         *standard_xauth_path = standard_xauthority_path();
    display_node *displays            = determine_displays();
    xauth_node   *xauth_list          = build_auth_list(standard_xauth_path);

    if (displays) {
        if (xauth_list) {
            ok = replace_auth_record_hostname(xauth_list, displays, old_hostname, new_hostname);
            if (ok)
                ok = commit_auth_list(standard_xauth_path, xauth_list);
        }

        /* Try additional path if we has the local-session display */
        if (displays->is_local) {
            char       *additional_xauth_path = additional_xauthority_path(displays->display_name);
            xauth_node *additional_xauth_list = build_auth_list(additional_xauth_path);

            if (additional_xauth_list) {
                int success = replace_auth_record_hostname(additional_xauth_list, displays, old_hostname, new_hostname);
                if (success)
                    success = commit_auth_list(additional_xauth_path, additional_xauth_list);

                ok = ok || success;
                free_auth_list(&additional_xauth_list);
            }

            free(additional_xauth_path);
        }
    }

    free_auth_list(&xauth_list);
    free_display_list(&displays);
    free(standard_xauth_path);

    print_debug("perform_auth_changes(): return %s\n", ok ? "true" : "false");
    return ok;
}

#if (__GLIBC__ == 2 && __GLIBC_MINOR__ >= 32) || (__GLIBC__ > 2)
#define STRERRORNAME(RC) strerrorname_np(RC)
#else
#define STRERRORNAME(RC) ""
#endif

#define fail_on_error(...)                                                                                             \
    do {                                                                                                               \
        int _rc = (__VA_ARGS__);                                                                                       \
        if (_rc < 0) {                                                                                                 \
            fprintf(stderr,                                                                                            \
                    __FILE__ ":%i: " #__VA_ARGS__ " failed with rc = %i (-%s)\n",                                      \
                    __LINE__,                                                                                          \
                    _rc,                                                                                               \
                    STRERRORNAME(-_rc));                                                                               \
            return _rc;                                                                                                \
        }                                                                                                              \
    } while (0)

/* Retrieves the hostname from the org.freedesktop.DBus.Properties.PropertiesChanged signal message.
 */
int message_handler(sd_bus_message *message, void *user_data, sd_bus_error *return_error) {
    (void)return_error;

    fail_on_error(sd_bus_message_skip(message, "s"));
    fail_on_error(sd_bus_message_enter_container(message, SD_BUS_TYPE_ARRAY, "{sv}"));

    char *hostname        = NULL;
    char *static_hostname = NULL;

    while (true) {
        int rc;
        fail_on_error(rc = sd_bus_message_enter_container(message, SD_BUS_TYPE_DICT_ENTRY, "sv"));
        if (rc == 0)
            break;

        char *key = NULL;
        fail_on_error(sd_bus_message_read_basic(message, 's', &key));

        char **hostname_dst = NULL;
        bool   is_hostname  = false;
        if (!strcmp("Hostname", key)) {
            hostname_dst = &hostname;
            is_hostname  = true;
        }
        else if (!strcmp("StaticHostname", key)) {
            hostname_dst = &static_hostname;
            is_hostname  = true;
        }

        if (is_hostname) {
            const char *contents = NULL;
            fail_on_error(sd_bus_message_peek_type(message, NULL, &contents));
            if (!strcmp(contents, "s")) {
                fail_on_error(sd_bus_message_enter_container(message, SD_BUS_TYPE_VARIANT, contents));
                fail_on_error(sd_bus_message_read_basic(message, contents[0], hostname_dst));
                fail_on_error(sd_bus_message_exit_container(message));
            }
        } else {
            sd_bus_message_skip(message, "v");
        }

        fail_on_error(sd_bus_message_exit_container(message));
    }

    /* StaticHostname takes precedence if it specified */
    if (static_hostname && strlen(static_hostname) > 0)
        hostname = static_hostname;

    if (hostname) {
        char *current_hostname = (char *)user_data;
        perform_auth_changes(current_hostname, hostname);
        strncpy(current_hostname, hostname, HOST_NAME_MAX);
    }

    return 0;
}

int epollhup_handler(sd_event_source *event_src, int fd, uint32_t revents, void *userdata) {
    (void)fd;
    (void)revents;

    /* Prevent XCloseDisplay after EPOLLHUP in connection fd */
    Display **display = (Display **)userdata;
    *display = NULL;
    return sd_event_exit(sd_event_source_get_event(event_src), 0);
}

int run() {
    autocleanup(sd_event) *event = NULL;
    fail_on_error(sd_event_default(&event));

    /* Block signals */
    sigset_t sigset;
    if (sigemptyset(&sigset) < 0 || sigaddset(&sigset, SIGTERM) < 0 || sigaddset(&sigset, SIGINT) ||
        sigaddset(&sigset, SIGHUP) < 0) {
        fprintf(stderr, "failed to setup signal set\n");
        return EXIT_FAILURE;
    }
    fail_on_error(sigprocmask(SIG_BLOCK, &sigset, NULL));

    /* Exit event loop on SIGTERM, SIGINT and SIGHUP */
    fail_on_error(sd_event_add_signal(event, NULL, SIGTERM, NULL, NULL));
    fail_on_error(sd_event_add_signal(event, NULL, SIGINT, NULL, NULL));
    fail_on_error(sd_event_add_signal(event, NULL, SIGHUP, NULL, NULL));

    const char *display_name = getenv("DISPLAY");
    autocleanup(Display) *display = XOpenDisplay(display_name);
    if (!display) {
        print_error("Could not open display %s\n", (display_name ? display_name : "(NULL)"));
        return EXIT_FAILURE;
    }
    int x_fd = ConnectionNumber(display);
    /*
    if (!(fcntl(x_fd, O_NONBLOCK) & O_NONBLOCK)) {
        print_error("O_NONBLOCK is not set\n");
        return EXIT_FAILURE;
    }
    */

    autocleanup(sd_event_source) *event_source = NULL;
    /* Exit event loop on xcb connection shutdown */
    fail_on_error(sd_event_add_io(event, &event_source, x_fd, EPOLLHUP, epollhup_handler, &display));

    autocleanup(sd_bus) *bus = NULL;
    fail_on_error(sd_bus_default_system(&bus));

    /* Get current hostname */
    char *read_hostname = NULL;

    int rc = sd_bus_get_property_string(
        bus, HOSTNAME1_DEST, HOSTNAME1_PATH, HOSTNAME1_DEST, "StaticHostname", NULL, &read_hostname);
    if (rc == -EBADR) {
        rc = sd_bus_get_property_string(
            bus, HOSTNAME1_DEST, HOSTNAME1_PATH, HOSTNAME1_DEST, "Hostname", NULL, &read_hostname);
    }
    fail_on_error(rc);

    char current_hostname[HOST_NAME_MAX + 1];
    strncpy(current_hostname, read_hostname, HOST_NAME_MAX);
    free(read_hostname);

    print_debug("Get hostname: %s\n", current_hostname);

    fail_on_error(sd_bus_match_signal(bus,
                                      NULL,
                                      HOSTNAME1_DEST,
                                      HOSTNAME1_PATH,
                                      "org.freedesktop.DBus.Properties",
                                      "PropertiesChanged",
                                      &message_handler,
                                      current_hostname));
    fail_on_error(sd_bus_attach_event(bus, event, 0));
    fail_on_error(sd_event_loop(event));

    return 0;
}

void fork_follow_child() {
    pid_t pid = fork();

    if (pid < 0) {
        print_error("fork() failed");
        exit(EXIT_FAILURE);
    }

    /* Exit from parent */
    if (pid > 0)
        exit(EXIT_SUCCESS);
}

void daemonize() {
    fork_follow_child();

    /* Child */
    if (setsid() < 0)
        exit(EXIT_FAILURE);

    signal(SIGCHLD, SIG_IGN);
    signal(SIGHUP, SIG_IGN);

    fork_follow_child();

    /* Grandchild */
    umask(0);

    if (chdir("/")) {
        print_error("Could not change root dir\n");
        exit(EXIT_FAILURE);
    }

    CLOSE_ALL();
}

typedef struct options_s {
    bool run_as_daemon;
} options_t;

void parse_opts(options_t *options, int argc, char **argv) {
    int opt;
    while ((opt = getopt(argc, argv, "dh")) != -1) {
        switch (opt) {
        case 'd': options->run_as_daemon = true; break;
        default:
            fprintf(stderr, "Usage:\n  %s [-d]\n\nOptions:\n  -d\t\t\trun as daemon\n", argc > 0 ? argv[0] : "");
            exit(opt == 'h' ? EXIT_SUCCESS : EXIT_FAILURE);
        }
    }
}

int main(int argc, char **argv) {
    options_t opts = {.run_as_daemon = false};
    parse_opts(&opts, argc, argv);

    if (!(getenv("DISPLAY") && getenv("XAUTHORITY"))) {
        print_error("Environment variables XAUTHORITY/DISPLAY is not set\n");
        return EXIT_FAILURE;
    }

    if (opts.run_as_daemon)
        daemonize();

    return run() ? EXIT_FAILURE : EXIT_SUCCESS;
}
